/*
 * The SmartWeb Framework
 * Copyright (C) 2004-2006
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * For further informations on the SmartWeb Framework please visit
 *
 *                        http://smartweb.sourceforge.net
 */
 
package net.smartlab.web;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

/**
 * This interface should be implemented by classes which provide some form of
 * persistence.
 * 
 * @author rlogiacco
 */
public interface DataAccessObject {

	/**
	 * Retrieves from the persistence tier the object which primary key equals
	 * the one specified.
	 * 
	 * @param key the primary key used to search the instance into the
	 *            persistence tier.
	 * @return an object representing the datas stored in the persistence tier
	 *         associated with the specified key.
	 * @throws DAOException if an error occur while accessing the persistence
	 *             tier.
	 * @throws UndefinedKeyException if the specified primary key is not present
	 *             on the persistence tier.
	 */
	public Object findByKey(Serializable key) throws DAOException;

	/**
	 * Permanently deletes an instance from the persistence tier.
	 * 
	 * @param object the instance representing the informations to be deleted
	 *            from the store.
	 * @throws DAOException if an error occur while accessing the persistence
	 *             tier.
	 */
	public void remove(Object object) throws DAOException;

	/**
	 * Ensures the persistence tier representation of the object is consistent
	 * with the in memory representation. If the object doesn't exist yet into
	 * the persistence tier it must be added and a new primary key assigned to
	 * the object.
	 * 
	 * @param object the object to be persisted.
	 * @throws DAOException if an error occur while accessing the persistence
	 *             tier
	 */
	public void update(Object object) throws DAOException;

	/**
	 * Returns a collection of objects representing all the persistence tier
	 * informations matching the specified search criterias.
	 * 
	 * @param info the criterias to be used to search the persistence tier.
	 * @return a collection of matching entities.
	 * @throws DAOException if an error occur while accessing the persistence
	 *             tier
	 */
	public Collection list(SearchInfo info) throws DAOException;

	/**
	 * Instances of this class represents a set of criterias to be used in
	 * persistence tier searches.
	 * 
	 * @author rlogiacco
	 */
	public static class SearchInfo {

		/**
		 * Entity properties to be filtered.
		 */
		private Collection filters = new ArrayList();

		/**
		 * Property to be used for ordering.
		 */
		private String order = null;

		/**
		 * The ordering direction.
		 */
		private boolean descendant = false;
		
		/**
		 * The filtering style.
		 */
		private boolean union = false;

		/**
		 * Identifies a <code>greater</code> expression condition.
		 */
		public final static int EQUALS = 0;
		/**
		 * Identifies an <code>equals</code> expression condition.
		 */
		public final static int GREATER = 1;
		/**
		 * Identifies a <code>greater</code> expression condition.
		 */
		public final static int GREATER_EQUALS = 2;
		/**
		 * Identifies a <code>greater equals</code> expression condition.
		 */
		public final static int LESSER = 3;
		/**
		 * Identifies a <code>lesser</code> expression condition.
		 */
		public final static int LESSER_EQUALS = 4;
		/**
		 * Identifies a <code>lesser equal</code> expression condition.
		 */
		public final static int NOT_EQUALS = 5;
		/**
		 * Identifies a <code>like</code> expression condition.
		 */
		public final static int LIKE = 6;
		/**
		 * Identifies a <code>between</code> expression condition.
		 */
		public final static int BETWEEN = 7;

		/**
		 * @return returns the ordering direction.
		 */
		public boolean isDescendant() {
			return descendant;
		}

		/**
		 * @return returns the filters.
		 */
		public Collection getFilters() {
			return filters;
		}

		/**
		 * @param filters the filters to set.
		 */
		public void setFilters(Collection filters) {
			this.filters = filters;
		}

		/**
		 * @param filters the filters to set.
		 */
		public void setFilters(String[] filters) {
			if (filters != null) {
				for (int i = 0; i < filters.length; i++) {
					this.addFilter(filters[i]);
				}
			}
		}

		/**
		 * Adds a filter.
		 * 
		 * @param filter the filter to add.
		 */
		public void addFilter(String filter) {
			for (int i = 0; i < filter.length(); i++) {
				switch (filter.charAt(i)) {
					// Equal
					case '=':
						filters.add(new Filter(filter.substring(0, i), EQUALS, filter.substring(i + 1)));
						return;
					// Greater
					case '>':
						// Greater equal
						if (filter.charAt(i + 1) == '=') {
							filters.add(new Filter(filter.substring(0, i + 1), GREATER_EQUALS, filter.substring(i + 2)));
						} else {
							filters.add(new Filter(filter.substring(0, i), GREATER, filter.substring(i + 1)));
						}
						return;
					// Lesser
					case '<':
						// Lesser equal
						if (filter.charAt(i + 1) == '=') {
							filters.add(new Filter(filter.substring(0, i + 1), LESSER_EQUALS, filter.substring(i + 2)));
						} else {
							filters.add(new Filter(filter.substring(0, i), LESSER, filter.substring(i + 1)));
						}
						return;
					// Not equals
					case '!':
						filters.add(new Filter(filter.substring(0, i), NOT_EQUALS, filter.substring(i + 1)));
						return;
					// Between
					case '|':
						filters.add(new Filter(filter.substring(0, i), BETWEEN, filter.substring(i + 1)));
						return;
					// Like
					case '%':
						filters.add(new Filter(filter.substring(0, i), LIKE, filter.substring(i + 1)));
						return;
				}
			}
		}

		/**
		 * Adds a filter.
		 * 
		 * @param column the column to filter on.
		 * @param expression the expression condition to apply on the column.
		 * @param values a comma separated list of values to evaluate against
		 *            the condition.
		 */
		public void addFilter(String column, int expression, String values) {
			filters.add(new Filter(column, expression, values));
		}

		/**
		 * @return returns the order.
		 */
		public String getOrder() {
			return order;
		}

		/**
		 * Sets the property used to order the collection. By default the
		 * ordering is set to <code>descendant</code> unless an <b>! </b>
		 * <code>(exclamation mark)</code> is prefixed indicating an
		 * <code>ascendant</code> order must be used.
		 * 
		 * @param order the property used to order the collection, optionally
		 *            prefixed by <b>! </b> <code>(exclamation mark)</code> to
		 *            invert the sorting direction.
		 */
		public void setOrder(String order) {
			if (order != null && order.length() > 0) {
				if (order.charAt(0) == '!') {
					this.descendant = true;
					this.order = order.substring(1);
				} else {
					this.order = order;
				}
			} else {
				order = null;
			}
		}
		
		/**
		 * TODO documentation
		 * @param style
		 */
		public void setUnion(String style) {
			if (style != null && style.length() > 0) {
				if (style.equalsIgnoreCase("OR")) {
					this.union = true;
				} else {
					this.union = false;
				}
			}
		}
		
		/**
		 * TODO documentation
		 * @return
		 */
		public boolean isUnion() {
			return union;
		}

		/**
		 * Represents a filtering condition to be applied while performing
		 * searches.
		 * 
		 * @author rlogiacco
		 */
		protected class Filter {

			/**
			 * The column.
			 */
			private String column;

			/**
			 * The condition.
			 */
			private int condition;

			/**
			 * The values array.
			 */
			private String[] values;

			/**
			 * Creates a filter with a column, a condition and a comma separated
			 * list of values to check against.
			 * 
			 * @param column the column to filter on.
			 * @param condition the condition to apply.
			 * @param values a comma separated list of values to check against.
			 */
			private Filter(String column, int condition, String values) {
				this.column = column;
				this.condition = condition;
				List list = new ArrayList();
				for (int start = 0, end = values.indexOf(','); end > -1; start = end + 1, end = values.indexOf(start, ',')) {
					list.add(values.substring(start, end));
				}
				list.add(values.substring(values.lastIndexOf(',') + 1));
				this.values = (String[])list.toArray(new String[0]);
			}

			/**
			 * Returns the column.
			 * 
			 * @return the column.
			 */
			protected String getColumn() {
				return column;
			}

			/**
			 * Returns the condition.
			 * 
			 * @return the condition.
			 */
			protected int getCondition() {
				return condition;
			}

			/**
			 * Returns the values.
			 * 
			 * @return the values array.
			 */
			protected String[] getValues() {
				return values;
			}

			/**
			 * Returns the n-th value in the list where the specified index is 0
			 * based.
			 * 
			 * @param index the 0 based index of the value to retrieve.
			 * @return the n-th value in the values list.
			 */
			protected String getValue(int index) {
				return values[index];
			}
		}
	}
}